if ($host.Version.Major -LT 5) {
	throw "PowerShell version outdated. Please update. Required version: 5.1.x.x."
}

if ($host.Version.Minor -LT 1) {
	if ($host.Version.Major -LE 5) {
		throw "PowerShell version outdated. Please update. Required version: 5.1.x.x."
	}
}

Add-Type @"
using System;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
public class ServerCertificateValidationCallback {
    public static void DoNotIgnore(Object[] certificates) {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
                foreach(Object trustedCertificate in certificates) {
                    if (trustedCertificate.Equals(certificate)) {
                        return true;
                    }
                }
				return false;
			};
	}

	public static void Ignore() {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
				return true;
			};
	}

	public static void AcceptOnly(X509Certificate cert, Object[] certificates) {
		ServicePointManager.ServerCertificateValidationCallback += 
			delegate(Object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors errors) {
				if (cert.Equals(certificate)) {
                    return true;
                }

                foreach(Object trustedCertificate in certificates) {
                    if (trustedCertificate.Equals(certificate)) {
                        return true;
                    }
                }

                return false;
			};
	}
}
"@

$script:Version='4.4.6'
$script:APIVersion=0
$script:Address=''
$script:Port=443
$script:Credential=$null
$script:CertificateCheckDisabled=$false

# Disable Certificate check (at your own risk!)
# This will not only disable the certificate check for anynode requests, but for
# all requests.
function Disable-CertificateCheck {
    $script:CertificateCheckDisabled=$true
	[ServerCertificateValidationCallback]::Ignore()
}

function Enable-CertificateCheck {
    $script:CertificateCheckDisabled=$false
	[ServerCertificateValidationCallback]::DoNotIgnore((Get-ChildItem 'Cert:/CurrentUser/Root/'))
}

function Initialize-anynodeAddress {
    [CmdletBinding()]
    Param(
        [Parameter(mandatory=$true)]
        [string] $Address
    )

    if ($Address -eq $null -OR $Address -eq '') {
        echo 'Please enter an address.'
        return
    }

    $script:Address=$Address
    echo "Address to communicate with anynode is set to: $Address"
}

function Reset-anynodeAddress {
    $script:Address=''
    echo 'Reset address used to communicate with anynode.'
}

function Initialize-anynodePort {
    [CmdletBinding()]
    Param(
        [Parameter(mandatory=$true)]
        [string] $Port
    )

    if ($Port -eq $null -OR $Port -le 0) {
        echo 'Please enter a valid port.'
        return
    }

    $script:Port=$Port
    echo "Port to communicate with anynode is set to: $Port"
}

function Reset-anynodePort {
    $script:Port=443
    echo 'Reset port used to communicate with anynode to value: 443'
}

function Initialize-anynodeCredential {
    [CmdletBinding(DefaultParameterSetName='cred')]
    Param(
        [Parameter(mandatory=$true,ParameterSetName='cred')]
        [PSCredential] $Credential,
        [Parameter(mandatory=$true,ParameterSetName='userpass')]
        [string] $Username,
        [Parameter(mandatory=$true,ParameterSetName='userpass')]
        [string] $Password
    )

    if ($Credential -ne $null) {
        $script:Credential=$Credential
    } elseif ($Username -ne $null) {
        try {
            $SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
	        $script:Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
        } catch {
            echo 'Could not set default credential.'
            return
        }
    }

    echo 'Set default credential.'
}

function Reset-anynodeCredential {
    $script:Credential=$null
    echo 'Reset default credential.'
}

# During this session, you can only communicate with the given interface.
function Get-anynodeWebserverCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $CertificateFilePath
	)

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

    Disable-CertificateCheck

	try {
		$Request=[System.Net.WebRequest]::Create('https://'+$Address+':'+$Port+'/')
		$Request.GetResponse().Dispose()
		$Certificate=$Request.ServicePoint.Certificate

		if (!($CertificateFilePath -eq $null) -AND !($CertificateFilePath -eq '')) {
			$CertBytes=$Certificate.Export([System.Security.Cryptography.X509Certificates.X509ContentType]::Cert)
			Set-Content -Value $CertBytes -Encoding Byte -Path $CertificateFilePath
		}

		[ServerCertificateValidationCallback]::AcceptOnly($Certificate, (Get-ChildItem 'Cert:/CurrentUser/Root/'))
	} catch {
        Enable-CertificateCheck
		$_.Exception
		return
	}
}

function Import-anynodeWebserverCertificate {
	[CmdletBinding()]
	Param(
		[Parameter(mandatory=$true)]
		[string] $CertificatePath
	)

	try {
		$Certificate=[X509Certificate]::CreateFromCertFile($CertificatePath)
	} catch {
		$_.Exception
		return
	}

    [ServerCertificateValidationCallback]::AcceptOnly($Certificate, (Get-ChildItem 'Cert:/CurrentUser/Root/'))
    echo "Set certificate to $CertificatePath"
}

# Invokes
#	Download
function Invoke-anynodeDownload {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	try {
		$Request=Invoke-WebRequest -Method GET -URI $URI -Headers $Headers

		$Request
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                $Request=Invoke-WebRequest -Method GET -URI $URI -Headers $Headers

		        $Request
            } else {
                throw $_
            }
        } catch {
            $e=$_.Exception
		    $output=$e.InnerException

		    if ($output.Length -eq 0) {
			    $output=$e.Message+' '+$_.ToString()
		    }

		    $output
        }
	}
}

#	Request
function Invoke-anynodeRequest {
	[CmdletBinding()]
	Param(
		[Parameter(mandatory=$true)]
		[ValidateSet('Get','Put','Post','Delete')]
		[string] $Method,
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter()]
		[string] $Body,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	try {
		if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {
			$Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers -Body $Body

		} else {
			$Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers
		}

		$Request
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                if (($Method -eq 'PUT') -Or ($Method -eq 'Put') -Or ($Method -eq 'put') -Or ($Method -eq 'POST') -Or ($Method -eq 'Post') -Or ($Method -eq 'post')) {
			        $Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers -Body $Body

		        } else {
			        $Request=Invoke-RestMethod -Method $Method -URI $URI -Headers $Headers
		        }

		        $Request
            } else {
                throw $_
            }
        } catch {
            $e=$_.Exception
		    $output=$e.InnerException

		    if ($output.Length -eq 0) {
			    $output=$e.Message+' '+$_.ToString()
		    }

		    $output
        }
	}
}

#	Upload
function Invoke-anynodeUpload {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $URITail,
		[Parameter(mandatory=$true)]
		[Object[]] $Filenames,
		[Parameter()]
		[Hashtable] $Parameters,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	if ($APIVersion -eq 0) {
		$APIVersion=$script:APIVersion
	}

	if ($Address -eq '' -OR $Address -eq $null) {
        if ($script:Address -eq '') {
            $Address=Read-Host 'Address'
        } else {
            $Address=$script:Address
        }

        if ($Address -eq '') {
            echo 'Address required.'
            return
        }
	}

	if ($Port -eq 0) {
		if ($script:Port -eq 0) {
            $Port=Read-Host 'Port'
        } else {
            $Port=$script:Port
        }

        if ($Port -eq 0) {
            echo 'Port required.'
            return
        }
	}

	if (!($Credential -eq $null)) {
		if (!($Username -eq '') -Or !($Password -eq '')) {
			echo 'Use either Credentials OR Username and Password. Not both.'
			return
		}
	}

	if (!($Username -eq '')) {
		if ($Password -eq '') {
			echo 'To use Username and Password, both must be given.'
			return
		}

		$SecretPassword=ConvertTo-SecureString $Password -AsPlainText -Force
		$Credential=New-Object System.Management.Automation.PSCredential($Username,$SecretPassword)
	}

    if ($Credential -eq $null -AND $script:Credential -ne $null) {
        $Credential=$script:Credential
    }

	$URI='https://'+$Address+':'+$Port+'/api'

	$ParameterString='?version='+$APIVersion
	foreach ($Parameter in $Parameters.Keys) {
		$ParameterString+='&'+$Parameter+'='+$Parameters.$Parameter
	}

	if (!($BackendName -eq '')) {
		$ParameterString+='&backendName='+$BackendName
	}

	if (!($Credential -eq $null)) {
		$Username=$Credential.Username
		$Password=$Credential.GetNetworkCredential().Password

		$Base64String=[System.Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes($Username+':'+$Password))
	}

	$URI='https://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString

	if (!($Base64String -eq $null)) {
		$Headers=@{Authorization="Basic "+$Base64String}
	}

    $Base64String=$null

	$Boundary=[System.Guid]::NewGuid().ToString()
	$ContentType='multipart/form-data; boundary="'+$Boundary+'"'

	$NL="`r`n";

	$Body=[byte[]]::new(0)

	foreach ($Filename in $Filenames) {
		$Bytes=[System.IO.File]::ReadAllBytes($Filename)

		$Head=(
			"--$Boundary",
		    "Content-Disposition: form-data; name=`"file`"; filename=`"$Filename`"",
		    "Content-Type: application/octet-stream$NL$NL"
		) -join $NL

		$Body+=[Text.Encoding]::UTF8.GetBytes($Head)
		$Body+=$Bytes

		$Foot="$NL--$Boundary--$NL"
		$Body+=[Text.Encoding]::UTF8.GetBytes($Foot)
	}

	$tmp=New-TemporaryFile
	$tmpName=$tmp.FullName
	[IO.File]::WriteAllBytes($tmpName,$Body)

	try {
		Invoke-RestMethod -Method PUT -Uri $URI -Headers $Headers -ContentType $ContentType -InFile $tmpName
	} catch {
        try {
            if ($script:CertificateCheckDisabled) {
                $URI='http://'+$Address+':'+$Port+'/api'+$URITail+$ParameterString
                Invoke-RestMethod -Method PUT -Uri $URI -Headers $Headers -ContentType $ContentType -InFile $tmpName

		        $Request
            } else {
                throw $_
            }
        } catch {
            $e=$_.Exception
		    $output=$e.InnerException

		    if ($output.Length -eq 0) {
			    $output=$e.Message+' '+$_.ToString()
		    }

		    $output
        }
	} finally {
		Remove-Item $tmpName -Force	
	}
}

# Active Sessions
#	Get
function Get-anynodeActiveSessions {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /activeSessions/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Backends
#	Get
function Get-anynodeBackends {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /backends/get -Credential $Credential -APIVersion $APIVersion -Username $Username -Password $Password
}

# Config
#	Get
function Get-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Filename -eq '')) {
		$Parameters.Add("filename",$Filename)
	}

	$Request=Invoke-anynodeDownload -Address $Address -Port $Port -URITail /anynodeConfig/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	if ($Filename -eq '') {
		if ($Request.Headers -eq $null) {
			echo $Request
			return
		}

		$ContentDisposition=$Request.Headers."Content-Disposition"
		if ($ContentDisposition -eq $null) {
			echo 'Could not find Content-Disposition header.'
			return
		}

		$split=$ContentDisposition -split 'filename="'
		if (($split[1] -eq $null) -Or ($split[1] -eq '')) {
			echo 'Could not find filename.'
			return
		}

		$Filename=($split[1] -split '"')[0]
	}

	try {
		[Text.Encoding]::Default.GetString($Request.Content) | Out-File $Filename -Encoding default
		echo 'Saved file at "'$Filename'".'
	} catch {
		echo $_.Exception
	}
}

#	Import
function Import-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[switch] $CommitProtection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=''
	try {
		$Body=[IO.File]::ReadAllText($Filepath)
	} catch {
		echo 'Could not find file.'
		return
	}

	$Parameters=@{}
	if ($CommitProtection) {
		$Parameters.Add("commitProtection",$true)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /anynodeConfig/set -Filenames @($Filepath) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Push
function Push-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /anynodeConfig/commit -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Reset
function Reset-anynodeConfig {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /anynodeConfig/revert -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Dashboard
#	Get
function Get-anynodeDashboard {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('sipNodes','sfbUcmaNodes','ldapConnections')]
		[psobject] $List,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($List -eq $null)) {
		$Parameters.Add("list",$List)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /dashboard/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# License
#	Get
function Get-anynodeLicenses {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /licenses/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeLicenses {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /licenses/clear -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import
function Import-anynodeLicense {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=''
	try {
		$Body=[IO.File]::ReadAllText($Filepath)
	} catch {
		echo 'Could not find file.'
		return
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /licenses/importFile -Filenames @($Filepath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# LOD
#	Import
function Import-anynodeLOD {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $LODKey,
		[Parameter()]
		[string] $LicensingDevice,
		[Parameter(mandatory=$true)]
		[string] $EmailAddress,
		[Parameter(mandatory=$true)]
		[string] $CompanyName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	$Body.Add("lodKey",$LODKey)

	if (!($LicensingDevice -eq '')) {
		$Body.Add("licensingDevice",$LicensingDevice)
	}

	$Body.Add("emailAddress",$EmailAddress)

	$Body.Add("companyName",$CompanyName)

	$license=Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /licenses/importLod -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	$Identifier=$license.identifier
	
	if ($Identifier -eq $null) {
		$license
		return
	}

	$LicensesLoaded=Get-anynodeLicenses -Address $Address -Port $Port -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	$Result

	foreach ($LicenseLoaded in $LicensesLoaded) {
		if ($LicenseLoaded.identifier -eq $Identifier) {
			$Result=$LicenseLoaded
		}
	}

	if ($Result -eq $null) {
		$license
	} else {
		$Result
	}
}

# Maintenance Mode
#	Get
function Get-anynodeMaintenanceMode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /maintenanceMode/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeMaintenanceMode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ModeActive,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	
	if (!($ModeActive -eq $null)) {
		$Body.Add("modeActive",$ModeActive)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /maintenanceMode/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Module Version
#	Get
function Get-anynodeModuleVersion {
	$script:Version
}

# Network Controller
#	Get
function Get-anynodeNetworkControllers {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /networkControllers/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Interfaces
function Get-anynodeNetworkInterfaces {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /networkInterfaces/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node
#	Get
function Get-anynodeNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $IncludeNetworkInformation,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}

    if ($IncludeNetworkInformation -ne $null) {
        $Parameters.Add("includeNetworkInformation",$IncludeNetworkInformation)
    }

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Provider Profiles
function Get-anynodeProviderNodeProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/providerProfiles/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Provider
function Add-anynodeProviderNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[psobject] $DeutscheTelekomNumberExtensionRange,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewDeutscheTelekomSipTrunk=$true,
		[Parameter()]
		[string] $DeutscheTelekomSipTrunkId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateNatTraversal,
		[Parameter()]
		[string] $ExternalHost,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NatDeviceHasFixedPortMapping,
		[Parameter()]
		[int] $FixedUdpTcpPort,
		[Parameter()]
		[int] $FixedTlsPort,
		[Parameter()]
		[int] $FixedUdpPortRangeStart,
		[Parameter()]
		[int] $FixedTcpPortRangeStart,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $NodeViaSipTrunking,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[string] $ProxyUri,
		[Parameter()]
		[string] $PrimaryProxyUri,
		[Parameter()]
		[string] $SecondaryProxyUri,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $DefineAssertedUri,
		[Parameter()]
		[string] $AssertedUri,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeProxyInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

	if (!($DeutscheTelekomNumberExtensionRange -eq $null)) {
		$Body.Add("deutscheTelekomNumberExtensionRange",$DeutscheTelekomNumberExtensionRange)
	}

	$Body.Add("createNewDeutscheTelekomSipTrunk",$CreateNewDeutscheTelekomSipTrunk)

	if (!($DeutscheTelekomSipTrunkId -eq $null) -AND !($DeutscheTelekomSipTrunkId -eq '')) {
		$Body.Add("deutscheTelekomSipTrunkId",$DeutscheTelekomSipTrunkId)
	}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($ActivateNatTraversal -eq $null)) {
		$Body.Add("activateNatTraversal",$ActivateNatTraversal)
	}
	if ($ActivateNatTraversal) {
		$Body.Add("externalHost",$ExternalHost)

		if (!($NatDeviceHasFixedPortMapping -eq $null)) {
			$Body.Add("natDeviceHasFixedPortMapping",$NatDeviceHasFixedPortMapping)
		}
		if ($NatDeviceHasFixedPortMapping) {
			$Body.Add("fixedUdpTcpPort",$FixedUdpTcpPort)
			$Body.Add("fixedTlsPort",$FixedTlsPort)
			$Body.Add("fixedUdpPortRangeStart",$FixedUdpPortRangeStart)
			$Body.Add("fixedTcpPortRangeStart",$FixedTcpPortRangeStart)
		}
	}

	if (!($NodeViaSipTrunking -eq $null)) {
		$Body.Add("nodeViaSipTrunking",$NodeViaSipTrunking)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}

	if (!($ClientCredentials -eq $null)) {
		$Body.Add("credentials",$ClientCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($ProxyUri -eq $null) -AND !($ProxyUri -eq '')) {
		$Body.Add("proxyUri",$ProxyUri)
	}

	if (!($PrimaryProxyUri -eq $null) -AND !($PrimaryProxyUri -eq '')) {
		$Body.Add("primaryProxyUri",$PrimaryProxyUri)
	}

	if (!($SecondaryProxyUri -eq $null) -AND !($SecondaryProxyUri -eq '')) {
		$Body.Add("secondaryProxyUri",$SecondaryProxyUri)
	}

	if (!($DefineAssertedUri -eq $null)) {
		$Body.Add("defineAssertedUri",$DefineAssertedUri)
	}
	if ($DefineAssertedUri) {
		$Body.Add("assertedUri",$AssertedUri)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addProvider -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get System Profiles
function Get-anynodeSystemNodeProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($CompleteData) {
		$Parameters.Add("completeData",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/systemProfiles/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add System
function Add-anynodeSystemNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Profile,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[ValidateSet("REGISTRAR","REGISTRATION","TRUNK")]
		[psobject] $SipInterconnection,
		[Parameter()]
		[psobject] $RemoteSipDomain,
		[Parameter()]
		[string] $LocalSipDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateClientAuthentication,
		[Parameter()]
		[psobject] $ClientCredentials,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ActivateServerAuthentication,
		[Parameter()]
		[psobject] $ServerCredentials,
		[Parameter()]
		[string] $RegistrarUri,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[psobject] $SipRegistrar,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $UseNetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeRemoteSipDomainInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeSipRegistrarInWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeOwnSubnetInWhitelist,
		[Parameter()]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $AllowOnlyNegotiatedPeersForRtpRtcp,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=2,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{profile=$Profile}

	$Body=@{}

	$Body.Add("createNewNetworkController",$CreateNewNetworkController)
	if ($CreateNewNetworkController) {
		if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
			$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
		}
		$Body.Add("networkControllerInterface",$NetworkControllerInterface)
		$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
		$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
		if (!($EnableReverseDnsLookup -eq $null)) {
			$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
		}
	}

	if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
		$Body.Add("networkControllerId",$NetworkControllerId)
	}

	if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
		$Body.Add("udpTcpPort",$UdpTcpPort)
	}

	if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
		$Body.Add("tlsPort",$TlsPort)
	}

	if (!($UdpPortRange -eq $null)) {
		$Body.Add("udpPortRange",$UdpPortRange)
	}

	if (!($TcpPortRange -eq $null)) {
		$Body.Add("tcpPortRange",$TcpPortRange)
	}

	if (!($SipInterconnection -eq $null)) {
		$Body.Add("sipInterconnection",$SipInterconnection)
	}

	if (!($RemoteSipDomain -eq $null)) {
		$Body.Add("remoteSipDomain",$RemoteSipDomain)
	}

	if (!($LocalSipDomain -eq $null) -AND !($LocalSipDomain -eq '')) {
		$Body.Add("localSipDomain",$LocalSipDomain)
	}

	if (!($ActivateClientAuthentication -eq $null)) {
		$Body.Add("activateClientAuthentication",$ActivateClientAuthentication)
	}
	if (!($ClientCredentials -eq $null)) {
		$Body.Add("clientCredentials",$ClientCredentials)
	}

	if (!($ActivateServerAuthentication -eq $null)) {
		$Body.Add("activateServerAuthentication",$ActivateServerAuthentication)
	}
	if (!($ServerCredentials -eq $null)) {
		$Body.Add("serverCredentials",$ServerCredentials)
	}

	if (!($RegistrarUri -eq $null) -AND !($RegistrarUri -eq '')) {
		$Body.Add("registrarUri",$RegistrarUri)
	}

	if (!($AddressOfRecord -eq $null)) {
		$Body.Add("addressOfRecord",$AddressOfRecord)
	}

	if (!($SipRegistrar -eq $null)) {
		$Body.Add("sipRegistrar",$SipRegistrar)
	}

	if (!($UseNetworkPeerWhitelist -eq $null)) {
		$Body.Add("useNetworkPeerWhitelist",$UseNetworkPeerWhitelist)
	}

	$Body.Add("includeRemoteSipDomainInWhitelist",$IncludeRemoteSipDomainInWhitelist)
	$Body.Add("includeSipRegistrarInWhitelist",$IncludeSipRegistrarInWhitelist)
	$Body.Add("includeProxyInWhitelist",$IncludeProxyInWhitelist)
	$Body.Add("includeOwnSubnetInWhitelist",$IncludeOwnSubnetInWhitelist)
	$Body.Add("networkPeerWhitelist",$NetworkPeerWhitelist)
	$Body.Add("allowOnlyNegotiatedPeersForRtpRtcp",$AllowOnlyNegotiatedPeersForRtpRtcp)

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add MS Teams Tenant
function Add-anynodeMsTeamsTenantNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewTeamsCarrierNode=$false,
		[Parameter()]
		[string] $TeamsCarrierNodeId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $CreateNewNetworkController=$true,
		[Parameter()]
		[string] $NetworkControllerDisplayName,
		[Parameter()]
		[string] $NetworkControllerInterface='[Any Interface]',
		[Parameter()]
		[ValidateSet('[Any Address Version]','IN_ADDRESS_VERSION_4','IN_ADDRESS_VERSION_6')]
		[string] $NetworkControllerIpVersion='[Any Address Version]',
		[Parameter()]
		[string] $NetworkControllerIpAddress='[Any IP Address]',
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $EnableReverseDnsLookup,
		[Parameter()]
		[string] $NetworkControllerId,
		[Parameter()]
		[int] $UdpTcpPort,
		[Parameter()]
		[int] $TlsPort,
		[Parameter()]
		[psobject] $UdpPortRange,
		[Parameter()]
		[psobject] $TcpPortRange,
		[Parameter()]
		[string] $SbcFqdn,
		[Parameter()]
		[string] $TenantFqdn,
		[Parameter()]
		[psobject] $IncomingManipulations,
		[Parameter()]
		[psobject] $OutgoingManipulations,
		[Parameter()]
		[psobject] $RoutingDomains,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=2,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Profile=Get-anynodeSystemNodeProfiles -Address $Address -Port $Port -Credential $Credential -Username $Username -Password $Password | Where-Object -Property "displayName" -EQ "Microsoft Teams Direct Routing Carrier"

	$Parameters=@{profile=$Profile.id}

	$Body=@{}

	$Body.Add("createNewTeamsCarrierNode",$CreateNewTeamsCarrierNode)
	if ($CreateNewTeamsCarrierNode) {

		$Body.Add("createNewNetworkController",$CreateNewNetworkController)
		if ($CreateNewNetworkController) {
			if (!($NetworkControllerDisplayName -eq $null) -AND !($NetworkControllerDisplayName -eq '')) {
				$Body.Add("networkControllerDisplayName",$NetworkControllerDisplayName)
			}
			$Body.Add("networkControllerInterface",$NetworkControllerInterface)
			$Body.Add("networkControllerIpVersion",$NetworkControllerIpVersion)
			$Body.Add("networkControllerIpAddress",$NetworkControllerIpAddress)
			if (!($EnableReverseDnsLookup -eq $null)) {
				$Body.Add("enableReverseDnsLookup",$EnableReverseDnsLookup)
			}
		}

		if (!($NetworkControllerId -eq $null) -AND !($NetworkControllerId -eq '')) {
			$Body.Add("networkControllerId",$NetworkControllerId)
		}
	
		if (!($UdpTcpPort -eq $null) -AND $UdpTcpPort -gt 0) {
			$Body.Add("udpTcpPort",$UdpTcpPort)
		}
	
		if (!($TlsPort -eq $null) -AND $TlsPort -gt 0) {
			$Body.Add("tlsPort",$TlsPort)
		}
	
		if (!($UdpPortRange -eq $null)) {
			$Body.Add("udpPortRange",$UdpPortRange)
		}
	
		if (!($TcpPortRange -eq $null)) {
			$Body.Add("tcpPortRange",$TcpPortRange)
		}

		if (!($SbcFqdn -eq $null) -AND !($SbcFqdn -eq '')) {
			$Body.Add("sbcFqdn",$SbcFqdn)
		}
	}

	if (!($TeamsCarrierNodeId -eq $null) -AND !($TeamsCarrierNodeId -eq '')) {
		$Body.Add("teamsCarrierNodeId",$TeamsCarrierNodeId)
	}

	if (!($TenantFqdn -eq $null) -AND !($TenantFqdn -eq '')) {
		$Body.Add("tenantFqdn",$TenantFqdn)
	}

	if (!($IncomingManipulations -eq $null)) {
		if ($IncomingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("incomingManipulations",$IncomingManipulations)
		} else {
			$Body.Add("incomingManipulations",@{0=$IncomingManipulations}.Values)
		}
	}

	if (!($OutgoingManipulations -eq $null)) {
		if ($OutgoingManipulations.GetType().Name -eq 'Object[]') {
			$Body.Add("outgoingManipulations",$OutgoingManipulations)
		} else {
			$Body.Add("outgoingManipulations",@{0=$OutgoingManipulations}.Values)
		}
	}

	if (!($RoutingDomains -eq $null)) {
		$Body.Add("routingDomains",$RoutingDomains)
	}

	if (!($DisplayName -eq $null) -AND !($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/addSystem -Body ($Body | ConvertTo-Json -Depth 99) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Incoming Manipulations
function Get-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/incomingManipulations/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Incoming Manipulations
function Set-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/incomingManipulations/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Incoming Manipulations
function Add-anynodeNodeIncomingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/incomingManipulations/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Incoming Manipulation
function Remove-anynodeNodeIncomingManipulation {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/incomingManipulations/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Outgoing Manipulations
function Add-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Outgoing Manipulation
function Remove-anynodeNodeOutgoingManipulation {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Outgoing Manipulations
function Get-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Outgoing Manipulations
function Set-anynodeNodeOutgoingManipulations {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[psobject] $Manipulations,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($Manipulations | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/outgoingManipulations/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Network Peer Whitelist
function Get-anynodeNodeNetworkPeerWhitelist {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Network Peer Whitelist
function Set-anynodeNodeNetworkPeerWhitelist {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($NetworkPeerWhitelist | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/set -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add Network Peer Whitelist Entry
function Add-anynodeNodeNetworkPeerWhitelistEntry {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[psobject] $NetworkPeerWhitelist,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	$Body=($NetworkPeerWhitelist | ConvertTo-Json)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/add -Parameters $Parameters -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove Network Peer Whitelist Entry
function Remove-anynodeNodeNetworkPeerWhitelistEntry {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $Hostname,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($Hostname -ne $null -AND $Hostname -ne '') {
		$Parameters.Add("hostname",$Hostname)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/networkPeerWhitelist/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[switch] $Force,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if ($Force) {
		$Parameters.Add("force",$true)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get MS Teams Connectivities
function Get-anynodeMsTeamsCarriers {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/msTeamsCarriers/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Certificates
#	Get
function Get-anynodeNodeCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/certificates/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[string] $CertificateFilePath,
		[Parameter()]
		[string] $KeyFilePath,
		[Parameter()]
		[string] $CertificatePassword,
		[Parameter()]
		[ValidateSet("CERTIFICATES_AND_PRIVATE_KEY","CERTIFICATES_ONLY","PRIVATE_KEY_ONLY")]
		[string] $Mode='CERTIFICATES_AND_PRIVATE_KEY',
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if (!($CertificatePassword -eq '')) {
		$Parameters.Add("password",$CertificatePassword)
	}

	if (!($CertificateFilePath -eq '')) {
		$RequestParams=$Parameters.Clone()
		if ($Mode -eq 'CERTIFICATES_AND_PRIVATE_KEY') {
			$RequestParams.Add("mode",'CERTIFICATES_ONLY')
		} elseif (!($Mode -eq '')) {
			$RequestParams.Add("mode",$Mode)
		}

		Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $RequestParams -Filenames @($CertificateFilePath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
	}

	if (!($KeyFilePath -eq '')) {
		$RequestParams=$Parameters.Clone()
		if ($Mode -eq 'CERTIFICATES_AND_PRIVATE_KEY') {
			$RequestParams.Add("mode",'PRIVATE_KEY_ONLY')
		} elseif (!($Mode -eq '')) {
			$RequestParams.Add("mode",$Mode)
		}

		Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $RequestParams -Filenames @($KeyFilePath) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
	}

# This doesn't work because anynode only accepts one file at a time.
#	$Files=@{}
#	if (!($CertificateFilePath -eq '')) {
#		$Files.Add(1,$CertificateFilePath)
#	}
#
#	if (!($KeyFilePath -eq '')) {
#		$Files.Add(2,$KeyFilePath)
#	}
#
#	if ($Files.Count -eq 0) {
#		echo 'Path(s) to Certificate and/or Private Key must be given for this call to work.'
#		return
#	}
#
#	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/certificates/set -Parameters $Parameters -Filenames $Files.Values -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Trusted Certificates
#	Get
function Get-anynodeNodeTrustedCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/trustedCertificates/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeNodeTrustedCertificate {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $CertificateFilePath,
		[Parameter()]
		[string] $CertificatePassword,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{node=$Node}
	if (!($CertificatePassword -eq '')) {
		$Parameters.Add("password",$CertificatePassword)
	}

	$Files=@{}
	if (!($CertificateFilePath -eq '')) {
		$Files.Add(1,$CertificateFilePath)
	}

	if ($Files.Count -eq 0) {
		echo 'Path to a Certificate must be given for this call to work.'
		return
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /nodes/trustedCertificates/add -Parameters $Parameters -Filenames $Files.Values -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeNodeTrustedCertificates {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/trustedCertificates/clear -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node FQDN
#	Get
function Get-anynodeNodeFQDN {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/tenants/fqdn/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeFQDN {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $NewFQDN,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/tenants/fqdn/set -Body $NewFQDN -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node override hostname
#	Get
function Get-anynodeNodeOverrideHostname {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/overrideHostname/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeOverrideHostname {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[string] $NewOverrideHostname,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/overrideHostname/set -Body $NewOverrideHostname -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Port
#	Get
function Get-anynodeNodePort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/port/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodePort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[int] $NewPort,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/port/set -Body $NewPort -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node TLS Port
#	Get
function Get-anynodeNodeTlsPort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/tlsPort/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeNodeTlsPort {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[int] $NewPort,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /nodes/tlsPort/set -Body $NewPort -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Node Transport Connection
#	Get
function Get-anynodeNodeTransportConnections {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /nodes/transportConnections/get -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeNodeTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter(mandatory=$true)]
		[psobject]$MatchingCondition,
		[Parameter(mandatory=$true)]
		[psobject]$TransportConnectionId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add("transportConnectionId",$TransportConnectionId)
	$Body.Add("matchingCondition",$MatchingCondition)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /nodes/transportConnections/add -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Clear
function Clear-anynodeNodeTransportConnections {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Node,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /nodes/transportConnections/clear -Parameters @{node=$Node} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Transport Connections
#	Get
function Get-anynodeTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/get -Parameters @{id=$Id} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Authentication
function Get-anynodeTransportConnectionAuthentication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/authentication/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Authentication
function Set-anynodeTransportConnectionAuthentication {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $TransportConnectionUsername,
		[Parameter()]
		[string] $TransportConnectionPassword,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($TransportConnectionUsername -eq '')) {
		$Body.Add('username',$TransportConnectionUsername)
	}

	if (!($TransportConnectionPassword -eq '')) {
		$Body.Add('password',$TransportConnectionPassword)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/authentication/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Asserted URI
function Get-anynodeTransportConnectionAssertedURI {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/assertedUri/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Asserted URI
function Set-anynodeTransportConnectionAssertedURI {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $URI,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($URI -eq '')) {
		$Body.Add('iri',$URI)
	}

	if (!($DisplayName -eq '')) {
		$Body.Add('displayName',$DisplayName)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/assertedUri/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Get Address of Record
function Get-anynodeTransportConnectionAddressOfRecord {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /transportConnections/addressOfRecord/get -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set Address of Record
function Set-anynodeTransportConnectionAddressOfRecord {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $TransportConnection,
		[Parameter()]
		[string] $URI,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($URI -eq '')) {
		$Body.Add('iri',$URI)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/addressOfRecord/set -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{transportConnection=$TransportConnection} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
#		Standard
function Add-anynodeStandardTransportConnection {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[string] $SipNodeId,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $UseAProxyServer,
		[Parameter()]
		[psobject] $Proxy,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $RegistrationEnabled,
		[Parameter()]
		[string] $SipDomain,
		[Parameter()]
		[string] $AddressOfRecord,
		[Parameter()]
		[ValidateSet($true, $false)]
		[psobject] $AuthenticationEnabled,
		[Parameter()]
		[psobject] $AuthenticationPolicy,
		[Parameter()]
		[string] $AssertedURI,
		[Parameter()]
		[string] $AssertedDisplayName,
		[Parameter()]
		[string] $ConditionId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	if (!($SipNodeId -eq '')) {
		$Body.Add('sipNodeId',$SipNodeId)
	}

	if (!($Proxy -eq $null)) {
		$Body.Add('proxy',$Proxy)
	}

	if ($RegistrationEnabled) {
		$Registration=@{}
		if (!($SipDomain -eq '')) {
			$Registration.Add('sipDomain',$SipDomain)
		}

		if (!($AddressOfRecord -eq '')) {
			$Registration.Add('addressOfRecord',$AddressOfRecord)
		}

		$Body.Add('registration',$Registration)
	}

	if (!($AuthenticationPolicy -eq '')) {
		$Body.Add('credentials',$AuthenticationPolicy)
	}

	if (!($AssertedURI -eq '')) {
		$AssertedAddress=@{iri=$AssertedURI}

		if (!($AssertedDisplayName -eq '')) {
			$AssertedAddress.Add('displayName',$AssertedDisplayName)
		}

		$Body.Add('assertedAddress',$AssertedAddress)
	}

	if (!($ConditionId -eq '')) {
		$Body.Add('conditionId',$ConditionId)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /transportConnections/addStandard -Body ($Body | ConvertTo-Json -Depth 99) -Parameters @{displayName=$DisplayName} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Object
#	Get
#		Single
function Get-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#		Multiple
function Get-anynodeObjects {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Sort,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($Sort) {
		$Parameters.Add("sort",$Sort)
	}

	if ($DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	if ($Parameters.Sort -eq $null -AND $Parameters.DisplayName -eq $null) {
        do {
		    $p1=Read-Host 'Choose a filter: sort (s)/display name (d)/both (b)'
        } while ($p1 -eq $null)

        switch ($p1) {
            's'{
                $Parameters.Add('sort',(Read-Host 'Sort'))
            }
            'd' {
                $Parameters.Add('displayName',(Read-Host 'Display Name'))
            }
            'b' {
                $Parameters.Add('sort',(Read-Host 'Sort'))
                $Parameters.Add('displayName',(Read-Host 'Display Name'))
            }
            default {
                echo 'Invalid input. Needs to be "s" or "d" or "b".'
                return
            }
        }
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/sorts -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter(mandatory=$true)]
		[psobject] $Object,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=($Object | ConvertTo-Json -Depth 99)

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /objects/set -Body $Body -Parameters @{id=$Id} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /objects/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Import
function Import-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Filepath,
		[Parameter()]
		[ValidateSet("SKIP","OVERRIDE","CLONE","CANCEL")]
		[psobject] $DuplicateOption,
		[Parameter()]
		[string] $CloneComment,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $ReassignPortOnCollision,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($DuplicateOption -eq $null)) {
		$Parameters.Add("action",$DuplicateOption)
	}

	if (!($CloneComment -eq '')) {
		$Parameters.Add("cloneComment",$CloneComment)
	}

	if ($ReassignPortOnCollision -ne $null) {
		$Parameters.Add("reassignPortOnCollision",$ReassignPortOnCollision)
	}

	Invoke-anynodeUpload -Address $Address -Port $Port -URITail /objects/import -Filenames @($Filepath) -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Export
function Export-anynodeObject {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $Id,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludePrivateData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{id=$Id}
	if (!($IncludePrivateData -eq $null)) {
		$Parameters.Add("includePrivateData",$IncludePrivateData)
	}

	$Request=Invoke-anynodeDownload -Address $Address -Port $Port -URITail /objects/export -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	if ($Request.Content -eq $null) {
		$Request
		return
	}

	if ($Filename -eq '') {
		if ($Request.Headers -eq $null) {
			echo $Request
			return
		}

		$ContentDisposition=$Request.Headers."Content-Disposition"
		if ($ContentDisposition -eq $null) {
			echo 'Could not find Content-Disposition header.'
			return
		}

		$split=$ContentDisposition -split 'filename="'
		if (($split[1] -eq $null) -Or ($split[1] -eq '')) {
			echo 'Could not find filename.'
			return
		}

		$Filename=($split[1] -split '"')[0]
	}

	[Text.Encoding]::Default.GetString($Request.Content) | Out-File $Filename -Encoding default

	echo 'Saved file at "'$Filename'".'
}

# Routing Domain
#	Get
function Get-anynodeRoutingDomains {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingDomains/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeSimpleRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DisplayName,
		[Parameter(mandatory=$true)]
		[string] $Node1,
		[Parameter(mandatory=$true)]
		[string] $Node2,
		[Parameter()]
		[string] $PrefixTo1,
		[Parameter()]
		[string] $PrefixTo2,
		[Parameter()]
		[string] $RoutingForwardProfile,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=0,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{"node1"=$Node1}
	$Body.Add("node2",$Node2)

	if (!($PrefixTo1 -eq '')) {
		$Body.Add("prefixTo1",$PrefixTo1)
	}

	if (!($PrefixTo2 -eq '')) {
		$Body.Add("prefixTo2",$PrefixTo2)
	}

	if (!($RoutingForwardProfile -eq '')) {
		$Body.Add("routingForwardProfile",$RoutingForwardProfile)
	}

	if (!($DisplayName -eq '')) {
		$Body.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/add -Body ($Body | ConvertTo-Json) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Add-anynodeRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[Object[]] $SourceNodes,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion=1,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}
	$Body.Add("displayName",$DisplayName)
    $Body.Add("sourceNodes",$SourceNodes)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/add -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#   Remove
function Remove-anynodeRoutingDomain {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /routingDomains/remove -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password -Parameters @{routingDomain=$RoutingDomain}
}

# Source Nodes
#	Add
function Add-anynodeRoutingDomainSourceNodes {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[Object[]] $SourceNodeIds,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=($SourceNodeIds | ConvertTo-Json)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/sourceNodes/add -Parameters @{routingDomain=$RoutingDomain} -Body $Body -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeRoutingDomainSourceNode {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[string] $SourceNodeId,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	$Parameters.Add('routingDomain',$RoutingDomain)
	$Parameters.Add('nodeId',$SourceNodeId)

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /routingDomains/sourceNodes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Routes
#	Add
function Add-anynodeEstablishmentRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $UseRouteSupervision,
		[Parameter()]
		[string] $SupervisionId,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[Object[]] $Filters,
		[Parameter()]
		[ValidateSet('ROUTE','IGNORE','REJECT','REDIRECT','ESTABLISH_PARALLEL')]
		[string] $Type,
		[Parameter()]
		[string] $DestinationNodeId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $SkipIfNodeNotOperational,
		[Parameter()]
		[ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
		[string] $RoutingForwardProfileSelection,
		[Parameter()]
		[string] $RoutingForwardProfileDisplayName,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AdjustRoutingForwardProfileSettings,
		[Parameter()]
		[ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
		[string] $RoutingForwardProfileMode,
		[Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
        [ValidateSet('SUCCESS','DIAL_STRING','NETWORK_PERMISSION','NETWORK_CONGESTION','NETWORK_EQUIPMENT','USER_BUSY','USER_REDIRECTED','USER_NOT_RESPONDING','USER_NOT_SELECTED','USER_REJECTED','USER_TERMINATED','MEDIA_NEGOTIATION','ERROR','DOMAIN_SPECIFIC_0','DOMAIN_SPECIFIC_1','DOMAIN_SPECIFIC_2','DOMAIN_SPECIFIC_3','DOMAIN_SPECIFIC_4','DOMAIN_SPECIFIC_5','DOMAIN_SPECIFIC_6','DOMAIN_SPECIFIC_7','DOMAIN_SPECIFIC_8','DOMAIN_SPECIFIC_9','DOMAIN_SPECIFIC_10','DOMAIN_SPECIFIC_11','DOMAIN_SPECIFIC_12','DOMAIN_SPECIFIC_13','DOMAIN_SPECIFIC_14','DOMAIN_SPECIFIC_15','DOMAIN_SPECIFIC_16','DOMAIN_SPECIFIC_17','DOMAIN_SPECIFIC_18','DOMAIN_SPECIFIC_19')]
        [string] $RejectStatus,
        [Parameter()]
        [string] $ParallelCallDialStringDirectoryId,
        [Parameter()]
        [PSCustomObject] $SourceDialStringRewriting,
        [Parameter()]
        [PSCustomObject] $DestinationDialStringRewriting,
        [Parameter()]
        [Object[]] $ParallelCalls,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","establishment")

	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }
    
    $Establishment=@{}

    if (!($Type -eq $null)) {
        $Establishment.Add("type",$Type)
    }

    if (!($DestinationNodeId -eq $null)) {
        $Establishment.Add("destinationNodeId",$DestinationNodeId)
    }

    if (!($SkipIfNodeNotOperational -eq $null)) {
        $Establishment.Add("skipIfNodeNotOperational",$SkipIfNodeNotOperational)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $Establishment.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $Establishment.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $Establishment.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $Establishment.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $Establishment.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    if (!($RejectStatus -eq $null)) {
        $Establishment.Add("rejectStatus",$RejectStatus)
    }

    if (!($ParallelCallDialStringDirectoryId -eq $null)) {
        $Establishment.Add("parallelCallDialStringDirectoryId",$ParallelCallDialStringDirectoryId)
    }

    if (!($SourceDialStringRewriting -eq $null)) {
        $Establishment.Add("sourceDialStringRewriting",$SourceDialStringRewriting)
    }

    if (!($DestinationDialStringRewriting -eq $null)) {
        $Establishment.Add("destinationDialStringRewriting",$DestinationDialStringRewriting)
    }

    if (!($ParallelCalls -eq $null)) {
        $Establishment.Add("parallelCalls",$ParallelCalls)
    }

    $Body.Add("establishment",$Establishment)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/add -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Add-anynodePathReplacementRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseUnconditionalRouting,
		[Parameter()]
		[ValidateSet($true,$false)]
        [pscustomobject] $UseRouteSupervision,
        [Parameter()]
        [string] $SupervisionId,
        [Parameter()]
        [string] $DisplayName,
        [Parameter()]
        [Object[]] $Filters,
        [Parameter()]
        [ValidateSet('ACCEPT','REJECT')]
        [string] $Type,
        [Parameter()]
        [string] $DestinationNodeId,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $SkipIfNodeNotOperational,
        [Parameter()]
        [ValidateSet('CREATE_NEW','USE_REFERENCE','NONE')]
        [string] $RoutingForwardProfileSelection,
        [Parameter()]
        [string] $RoutingForwardProfileDisplayName,
        [Parameter()]
        [ValidateSet($true,$false)]
        [pscustomobject] $AdjustRoutingForwardProfileSettings,
        [Parameter()]
        [ValidateSet('PASSTHROUGH','TRANSCODING','BYPASS','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
        [string] $RoutingForwardProfileMode,
        [Parameter()]
        [string] $RoutingForwardProfileId,
        [Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

    $Parameters=@{}
    $Parameters.Add("routingDomain",$RoutingDomain)
    $Parameters.Add("mode","pathReplacement")

	$Body=@{}

    if (!($UseUnconditionalRouting -eq $null)) {
        $Body.Add("useUnconditionalRouting",$UseUnconditionalRouting)
    }

    if (!($UseRouteSupervision -eq $null)) {
        $Body.Add("useRouteSupervision",$UseRouteSupervision)
    }

    if (!($SupervisionId -eq $null)) {
        $Body.Add("supervisionId",$SupervisionId)
    }
    
    if (!($DisplayName -eq $null)) {
        $Body.Add("displayName",$DisplayName)
    }

    if (!($Filters -eq $null)) {
        $Body.Add("filters",$Filters)
    }

    $PathReplacement=@{}

    if (!($Type -eq $null)) {
        $PathReplacement.Add("type",$Type)
    }

    if (!($RoutingForwardProfileSelection -eq $null)) {
        $PathReplacement.Add("routingForwardProfileSelection",$RoutingForwardProfileSelection)
    }

    if (!($RoutingForwardProfileDisplayName -eq $null)) {
        $PathReplacement.Add("routingForwardProfileDisplayName",$RoutingForwardProfileDisplayName)
    }

    if (!($AdjustRoutingForwardProfileSettings -eq $null)) {
        $PathReplacement.Add("adjustRoutingForwardProfileSettings",$AdjustRoutingForwardProfileSettings)
    }

    if (!($RoutingForwardProfileMode -eq $null)) {
        $PathReplacement.Add("routingForwardProfileMode",$RoutingForwardProfileMode)
    }

    if (!($RoutingForwardProfileId -eq $null)) {
        $PathReplacement.Add("routingForwardProfileId",$RoutingForwardProfileId)
    }

    $Body.Add("pathReplacement",$PathReplacement)

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingDomains/routes/add -Parameters $Parameters -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Remove
function Remove-anynodeEstablishmentRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{routingDomain=$RoutingDomain}
	$Parameters.Add("mode","establishment")

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /routingDomains/routes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

function Remove-anynodePathReplacementRoute {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $RoutingDomain,
		[Parameter()]
		[int] $Index=-1,
		[Parameter()]
		[string] $DisplayName,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{routingDomain=$RoutingDomain}
	$Parameters.Add("mode","pathReplacement")

	if ($Index -gt -1) {
		$Parameters.Add("index",$Index)
	} elseif ($DisplayName -ne $null -AND $DisplayName -ne '') {
		$Parameters.Add("displayName",$DisplayName)
	}

	Invoke-anynodeRequest -Method Delete -Address $Address -Port $Port -URITail /routingDomains/routes/remove -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Routing Forward Profile
#	Get
function Get-anynodeRoutingForwardProfiles {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $CompleteData,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /routingForwardProfiles/get -Parameters @{completeData=$CompleteData} -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Add
function Add-anynodeRoutingForwardProfile {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter(mandatory=$true)]
		[string] $DisplayName,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RemoveAnynodeFromSignalingPath,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $EarlyMediaPassThrough,
		[Parameter()]
		[pscustomobject] $TelephonyForwardingStack,
		[Parameter()]
		[string] $TelephonyForwardingStackId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RecordIncoming,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $RecordOutgoing,
		[Parameter()]
		[string] $MediaRecorderId,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AnnouncementIncoming,
		[Parameter()]
		[ValidateSet($true,$false)]
		[pscustomobject] $AnnouncementOutgoing,
		[Parameter()]
		[string] $AnnouncementMediaSourceId,
		[Parameter()]
		[string] $MediaNegotiationForwarderId,
		[Parameter()]
		[PSCustomObject] $SignalingForwardingFromCallingToCalled,
		[Parameter()]
		[PSCustomObject] $SignalingForwardingFromCalledToCalling,
		[Parameter()]
		[string] $MediaTranscodingOptionsId,
		[Parameter()]
		[ValidateSet('TRANSCODING','BYPASS','PASSTHROUGH','PASSTHROUGH_WITH_FALLBACK_TO_BYPASS')]
		[string] $MediaMode,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	$Body.Add('displayName',$DisplayName)

	if (!($RemoveAnynodeFromSignalingPath -eq $null)) {
		$Body.Add('removeAnynodeFromSignalingPath',$RemoveAnynodeFromSignalingPath)
	}

	if (!($EarlyMediaPassThrough -eq $null)) {
		$Body.Add('earlyMediaPassThrough',$EarlyMediaPassThrough)
	}

	if (!($TelephonyForwardingStack -eq $null)) {
		$Body.Add('telephonyForwardingStack',$TelephonyForwardingStack)
	}

	if (!($TelephonyForwardingStackId -eq '')) {
		$Body.Add('telephonyForwardingStackId',$TelephonyForwardingStackId)
	}

	if (!($RecordIncoming -eq $null)) {
		$Body.Add('recordIncoming',$RecordIncoming)
	}

	if (!($RecordOutgoing -eq $null)) {
		$Body.Add('recordOutgoing',$RecordOutgoing)
	}

	if (!($MediaRecorderId -eq '')) {
		$Body.Add('mediaRecorderId',$MediaRecorderId)
	}

	if (!($AnnouncementIncoming -eq $null)) {
		$Body.Add('announcementIncoming',$AnnouncementIncoming)
	}

	if (!($AnnouncementOutgoing -eq $null)) {
		$Body.Add('announcementOutgoing',$AnnouncementOutgoing)
	}

	if (!($AnnouncementMediaSourceId -eq '')) {
		$Body.Add('announcementMediaSourceId',$AnnouncementMediaSourceId)
	}

	if (!($MediaNegotiationForwarderId -eq '')) {
		$Body.Add('mediaNegotiationForwarderId',$MediaNegotiationForwarderId)
	}

	if (!($SignalingForwardingFromCallingToCalled -eq $null)) {
		$Body.Add('signalingForwardingFromCallingToCalled',$SignalingForwardingFromCallingToCalled)
	}

	if (!($SignalingForwardingFromCalledToCalling -eq $null)) {
		$Body.Add('signalingForwardingFromCalledToCalling',$SignalingForwardingFromCalledToCalling)
	}

	if (!($MediaTranscodingOptionsId -eq '')) {
		$Body.Add('mediaTranscodingOptionsId',$MediaTranscodingOptionsId)
	}

	if (!($MediaMode -eq '')) {
		$Body.Add('mediaMode',$MediaMode)
	}

	Invoke-anynodeRequest -Method Post -Address $Address -Port $Port -URITail /routingForwardProfiles/add -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Sorts
#	Get
function Get-anynodeSorts {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /objects/sorts -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# SystemIds
#	Get
function Get-anynodeSystemIds {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /licenses/systemIds/get -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Trace
#	Get
function Get-anynodeTrace {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[string] $Filename,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Filename -eq '')) {
		$Parameters.Add("filename",$Filename)
	}

	$Request=Invoke-anynodeDownload -Address $Address -Port $Port -URITail /trace/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password

	if ($Request.Content -eq $null) {
		$Request
		return
	}

	if ($Filename -eq '') {
		if ($Request.Headers -eq $null) {
			echo $Request
			return
		}

		$ContentDisposition=$Request.Headers."Content-Disposition"
		if ($ContentDisposition -eq $null) {
			echo 'Could not find Content-Disposition header.'
			return
		}

		$split=$ContentDisposition -split 'filename="'
		if (($split[1] -eq $null) -Or ($split[1] -eq '')) {
			echo 'Could not find filename.'
			return
		}

		$Filename=($split[1] -split '"')[0]
	}

	[Text.Encoding]::Default.GetString($Request.Content) | Out-File $Filename -Encoding default

	echo 'Saved file at "'$Filename'".'
}

# Tracing
#	Get
function Get-anynodeTracing {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[switch] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if ($IncludeDefaults) {
		$Parameters.Add("includeDefaults",$true)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /tracing/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

#	Set
function Set-anynodeTracing {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('FILE','SPLIT')]
		[string] $Selection,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $TraceActive,
		[Parameter()]
		[pscustomobject] $Options,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludePrivateData,
		[Parameter()]
		[ValidateSet($true,$false)]
		[psobject] $IncludeDefaults,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Body=@{}

	if (!($Selection -eq '')) {
		$Body.Add("selection",$Selection)
	}

	if (!($TraceActive -eq $null)) {
		$Body.Add("traceActive",$TraceActive)
	}

	if (!($Options -eq $null)) {
		$Body.Add("options",($Options | ConvertTo-Json -Depth 99 | ConvertFrom-Json))
	}

	if (!($IncludePrivateData -eq $null)) {
		$Body.Add("includePrivateData",$IncludePrivateData)
	}

	if (!($IncludeDefaults -eq $null)) {
		$Body.Add("includeDefaults",$IncludeDefaults)
	}

	Invoke-anynodeRequest -Method Put -Address $Address -Port $Port -URITail /tracing/set -Body ($Body | ConvertTo-Json -Depth 99) -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# Version
#	Get
function Get-anynodeVersion {
	[CmdletBinding()]
	Param(
		[Parameter()]
		[string] $Address,
		[Parameter()]
		[int] $Port,
		[Parameter()]
		[ValidateSet('anynode','java','ucma','administration','monitor','frontend')]
		[string] $Product,
		[Parameter()]
		[PSCredential] $Credential,
		[Parameter()]
		[string] $BackendName,
		[Parameter()]
		[int] $APIVersion,
		[Parameter()]
		[string] $Username,
		[Parameter()]
		[string] $Password
	)

	$Parameters=@{}
	if (!($Product -eq '')) {
		$Parameters.Add("product",$Product)
	}

	Invoke-anynodeRequest -Method Get -Address $Address -Port $Port -URITail /version/get -Parameters $Parameters -Credential $Credential -BackendName $BackendName -APIVersion $APIVersion -Username $Username -Password $Password
}

# SIG # Begin signature block
# MIIc8gYJKoZIhvcNAQcCoIIc4zCCHN8CAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUpp8bo2bPt7SXKXK6Wd2QAGUb
# 4hugghgJMIIDXzCCAkegAwIBAgILBAAAAAABIVhTCKIwDQYJKoZIhvcNAQELBQAw
# TDEgMB4GA1UECxMXR2xvYmFsU2lnbiBSb290IENBIC0gUjMxEzARBgNVBAoTCkds
# b2JhbFNpZ24xEzARBgNVBAMTCkdsb2JhbFNpZ24wHhcNMDkwMzE4MTAwMDAwWhcN
# MjkwMzE4MTAwMDAwWjBMMSAwHgYDVQQLExdHbG9iYWxTaWduIFJvb3QgQ0EgLSBS
# MzETMBEGA1UEChMKR2xvYmFsU2lnbjETMBEGA1UEAxMKR2xvYmFsU2lnbjCCASIw
# DQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAMwldpB5BngiFvXAg7aEyiie/QV2
# EcWtiHL8RgJDx7KKnQRfJMsuS+FggkbhUqsMgUdwbN1k0ev1LKMPgj0MK66X17YU
# hhB5uzsTgHeMCOFJ0mpiLx9e+pZo34knlTifBtc+ycsmWQ1z3rDI6SYOgxXG71uL
# 0gRgykmmKPZpO/bLyCiR5Z2KYVc3rHQU3HTgOu5yLy6c+9C7v/U9AOEGM+iCK65T
# pjoWc4zdQQ4gOsC0p6Hpsk+QLjJg6VfLuQSSaGjlOCZgdbKfd/+RFO+uIEn8rUAV
# SNECMWEZXriX7613t2Saer9fwRPvm2L7DWzgVGkWqQPabumDk3F2xmmFghcCAwEA
# AaNCMEAwDgYDVR0PAQH/BAQDAgEGMA8GA1UdEwEB/wQFMAMBAf8wHQYDVR0OBBYE
# FI/wS3+oLkUkrk1Q+mOai97i3Ru8MA0GCSqGSIb3DQEBCwUAA4IBAQBLQNvAUKr+
# yAzv95ZURUm7lgAJQayzE4aGKAczymvmdLm6AC2upArT9fHxD4q/c2dKg8dEe3jg
# r25sbwMpjjM5RcOO5LlXbKr8EpbsU8Yt5CRsuZRj+9xTaGdWPoO4zzUhw8lo/s7a
# wlOqzJCK6fBdRoyV3XpYKBovHd7NADdBj+1EbddTKJd+82cEHhXXipa0095MJ6RM
# G3NzdvQXmcIfeg7jLQitChws/zyrVQ4PkX4268NXSb7hLi18YIvDQVETI53O9zJr
# lAGomecsMx86OyXShkDOOyyGeMlhLxS67ttVb9+E7gUJTb0o2HLO02JQZR7rkpeD
# MdmztcpHWD9fMIIEpzCCA4+gAwIBAgIOSBtqB6lCTB6q/vPN8Q8wDQYJKoZIhvcN
# AQELBQAwTDEgMB4GA1UECxMXR2xvYmFsU2lnbiBSb290IENBIC0gUjMxEzARBgNV
# BAoTCkdsb2JhbFNpZ24xEzARBgNVBAMTCkdsb2JhbFNpZ24wHhcNMTYwNjE1MDAw
# MDAwWhcNMjQwNjE1MDAwMDAwWjBuMQswCQYDVQQGEwJCRTEZMBcGA1UEChMQR2xv
# YmFsU2lnbiBudi1zYTFEMEIGA1UEAxM7R2xvYmFsU2lnbiBFeHRlbmRlZCBWYWxp
# ZGF0aW9uIENvZGVTaWduaW5nIENBIC0gU0hBMjU2IC0gRzMwggEiMA0GCSqGSIb3
# DQEBAQUAA4IBDwAwggEKAoIBAQDZt7olrZTzW75BBhxTygwQjFFBWTN5ZPRXneTV
# JcTsUIRYmHJ5QOIveNSS6iYOnq6VfPvE/XFE3YxftyOLXrv0/EvLIz3DdgP10YxF
# vHF1HYvSiYm+41E9xsiKsjE1B2659bpqDfQQn67VYkkoe+xXuqsyfLF90qJWBjbu
# sO/Qaq7qqx/WDZ98lvutcJktXZXwgNB5RuxVOszTOPsEB6gHdYKC4NB+d7iP69Io
# /K5tFGhBf3ZD10i6YETht3Lo0PAgA3va2rQGdceyA974lMZoj157npudNuDO0mvG
# xmvpFCK1cX62j1of2+du9EIQkGjmK0UQT3O6LNfFMWpy3WNzAgMBAAGjggFjMIIB
# XzAOBgNVHQ8BAf8EBAMCAQYwHQYDVR0lBBYwFAYIKwYBBQUHAwMGCCsGAQUFBwMJ
# MBIGA1UdEwEB/wQIMAYBAf8CAQAwHQYDVR0OBBYEFNwsWCwqbzUtn3mVqEhdxG0+
# U7+5MB8GA1UdIwQYMBaAFI/wS3+oLkUkrk1Q+mOai97i3Ru8MD4GCCsGAQUFBwEB
# BDIwMDAuBggrBgEFBQcwAYYiaHR0cDovL29jc3AyLmdsb2JhbHNpZ24uY29tL3Jv
# b3RyMzA2BgNVHR8ELzAtMCugKaAnhiVodHRwOi8vY3JsLmdsb2JhbHNpZ24uY29t
# L3Jvb3QtcjMuY3JsMGIGA1UdIARbMFkwCwYJKwYBBAGgMgECMAcGBWeBDAEDMEEG
# CSsGAQQBoDIBXzA0MDIGCCsGAQUFBwIBFiZodHRwczovL3d3dy5nbG9iYWxzaWdu
# LmNvbS9yZXBvc2l0b3J5LzANBgkqhkiG9w0BAQsFAAOCAQEAdgnEzC/Z7x5LqfhX
# 80A5IcpMPB2eKSsg1CtE0ojOGg0Fz4OBu+tpvDGNKsTHRMxgYJQcz6HhAiQOrVu+
# LMInHme36CgfMlHjOfOY37ifLosqtHsKA7y9NgSPydCcT6MCJ5mw8EXpNN/kOqO3
# BjfYbyp5kNTUTlhx7FOpYZj3OWngEpxXWHKGJymlHeUy8yuZl1q/K7A8tAbqDmTs
# t81lgCQXwtk39bEmEDVHe5oCulSiRZP/eb8ajMWftZ/feOdrUPFHlGlLJLjaBegM
# nU8G7EoxIH5PXYaELzWjzZzBhFcfH63A4qSx7ylrIZem1P7tAzew/PWNKrzchIPj
# 3sPnXzCCBP4wggPmoAMCAQICEA1CSuC+Ooj/YEAhzhQA8N0wDQYJKoZIhvcNAQEL
# BQAwcjELMAkGA1UEBhMCVVMxFTATBgNVBAoTDERpZ2lDZXJ0IEluYzEZMBcGA1UE
# CxMQd3d3LmRpZ2ljZXJ0LmNvbTExMC8GA1UEAxMoRGlnaUNlcnQgU0hBMiBBc3N1
# cmVkIElEIFRpbWVzdGFtcGluZyBDQTAeFw0yMTAxMDEwMDAwMDBaFw0zMTAxMDYw
# MDAwMDBaMEgxCzAJBgNVBAYTAlVTMRcwFQYDVQQKEw5EaWdpQ2VydCwgSW5jLjEg
# MB4GA1UEAxMXRGlnaUNlcnQgVGltZXN0YW1wIDIwMjEwggEiMA0GCSqGSIb3DQEB
# AQUAA4IBDwAwggEKAoIBAQDC5mGEZ8WK9Q0IpEXKY2tR1zoRQr0KdXVNlLQMULUm
# EP4dyG+RawyW5xpcSO9E5b+bYc0VkWJauP9nC5xj/TZqgfop+N0rcIXeAhjzeG28
# ffnHbQk9vmp2h+mKvfiEXR52yeTGdnY6U9HR01o2j8aj4S8bOrdh1nPsTm0zinxd
# RS1LsVDmQTo3VobckyON91Al6GTm3dOPL1e1hyDrDo4s1SPa9E14RuMDgzEpSlwM
# MYpKjIjF9zBa+RSvFV9sQ0kJ/SYjU/aNY+gaq1uxHTDCm2mCtNv8VlS8H6GHq756
# WwogL0sJyZWnjbL61mOLTqVyHO6fegFz+BnW/g1JhL0BAgMBAAGjggG4MIIBtDAO
# BgNVHQ8BAf8EBAMCB4AwDAYDVR0TAQH/BAIwADAWBgNVHSUBAf8EDDAKBggrBgEF
# BQcDCDBBBgNVHSAEOjA4MDYGCWCGSAGG/WwHATApMCcGCCsGAQUFBwIBFhtodHRw
# Oi8vd3d3LmRpZ2ljZXJ0LmNvbS9DUFMwHwYDVR0jBBgwFoAU9LbhIB3+Ka7S5GGl
# sqIlssgXNW4wHQYDVR0OBBYEFDZEho6kurBmvrwoLR1ENt3janq8MHEGA1UdHwRq
# MGgwMqAwoC6GLGh0dHA6Ly9jcmwzLmRpZ2ljZXJ0LmNvbS9zaGEyLWFzc3VyZWQt
# dHMuY3JsMDKgMKAuhixodHRwOi8vY3JsNC5kaWdpY2VydC5jb20vc2hhMi1hc3N1
# cmVkLXRzLmNybDCBhQYIKwYBBQUHAQEEeTB3MCQGCCsGAQUFBzABhhhodHRwOi8v
# b2NzcC5kaWdpY2VydC5jb20wTwYIKwYBBQUHMAKGQ2h0dHA6Ly9jYWNlcnRzLmRp
# Z2ljZXJ0LmNvbS9EaWdpQ2VydFNIQTJBc3N1cmVkSURUaW1lc3RhbXBpbmdDQS5j
# cnQwDQYJKoZIhvcNAQELBQADggEBAEgc3LXpmiO85xrnIA6OZ0b9QnJRdAojR6Or
# ktIlxHBZvhSg5SeBpU0UFRkHefDRBMOG2Tu9/kQCZk3taaQP9rhwz2Lo9VFKeHk2
# eie38+dSn5On7UOee+e03UEiifuHokYDTvz0/rdkd2NfI1Jpg4L6GlPtkMyNoRdz
# DfTzZTlwS/Oc1np72gy8PTLQG8v1Yfx1CAB2vIEO+MDhXM/EEXLnG2RJ2CKadRVC
# 9S0yOIHa9GCiurRS+1zgYSQlT7LfySmoc0NR2r1j1h9bm/cuG08THfdKDXF+l7f0
# P4TrweOjSaH6zqe/Vs+6WXZhiV9+p7SOZ3j5NpjhyyjaW4emii8wggUxMIIEGaAD
# AgECAhAKoSXW1jIbfkHkBdo2l8IVMA0GCSqGSIb3DQEBCwUAMGUxCzAJBgNVBAYT
# AlVTMRUwEwYDVQQKEwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2Vy
# dC5jb20xJDAiBgNVBAMTG0RpZ2lDZXJ0IEFzc3VyZWQgSUQgUm9vdCBDQTAeFw0x
# NjAxMDcxMjAwMDBaFw0zMTAxMDcxMjAwMDBaMHIxCzAJBgNVBAYTAlVTMRUwEwYD
# VQQKEwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5jb20xMTAv
# BgNVBAMTKERpZ2lDZXJ0IFNIQTIgQXNzdXJlZCBJRCBUaW1lc3RhbXBpbmcgQ0Ew
# ggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC90DLuS82Pf92puoKZxTlU
# KFe2I0rEDgdFM1EQfdD5fU1ofue2oPSNs4jkl79jIZCYvxO8V9PD4X4I1moUADj3
# Lh477sym9jJZ/l9lP+Cb6+NGRwYaVX4LJ37AovWg4N4iPw7/fpX786O6Ij4YrBHk
# 8JkDbTuFfAnT7l3ImgtU46gJcWvgzyIQD3XPcXJOCq3fQDpct1HhoXkUxk0kIzBd
# vOw8YGqsLwfM/fDqR9mIUF79Zm5WYScpiYRR5oLnRlD9lCosp+R1PrqYD4R/nzEU
# 1q3V8mTLex4F0IQZchfxFwbvPc3WTe8GQv2iUypPhR3EHTyvz9qsEPXdrKzpVv+T
# AgMBAAGjggHOMIIByjAdBgNVHQ4EFgQU9LbhIB3+Ka7S5GGlsqIlssgXNW4wHwYD
# VR0jBBgwFoAUReuir/SSy4IxLVGLp6chnfNtyA8wEgYDVR0TAQH/BAgwBgEB/wIB
# ADAOBgNVHQ8BAf8EBAMCAYYwEwYDVR0lBAwwCgYIKwYBBQUHAwgweQYIKwYBBQUH
# AQEEbTBrMCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5kaWdpY2VydC5jb20wQwYI
# KwYBBQUHMAKGN2h0dHA6Ly9jYWNlcnRzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFz
# c3VyZWRJRFJvb3RDQS5jcnQwgYEGA1UdHwR6MHgwOqA4oDaGNGh0dHA6Ly9jcmw0
# LmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RDQS5jcmwwOqA4oDaG
# NGh0dHA6Ly9jcmwzLmRpZ2ljZXJ0LmNvbS9EaWdpQ2VydEFzc3VyZWRJRFJvb3RD
# QS5jcmwwUAYDVR0gBEkwRzA4BgpghkgBhv1sAAIEMCowKAYIKwYBBQUHAgEWHGh0
# dHBzOi8vd3d3LmRpZ2ljZXJ0LmNvbS9DUFMwCwYJYIZIAYb9bAcBMA0GCSqGSIb3
# DQEBCwUAA4IBAQBxlRLpUYdWac3v3dp8qmN6s3jPBjdAhO9LhL/KzwMC/cWnww4g
# Qiyvd/MrHwwhWiq3BTQdaq6Z+CeiZr8JqmDfdqQ6kw/4stHYfBli6F6CJR7Euhx7
# LCHi1lssFDVDBGiy23UC4HLHmNY8ZOUfSBAYX4k4YU1iRiSHY4yRUiyvKYnleB/W
# CxSlgNcSR3CzddWThZN+tpJn+1Nhiaj1a5bA9FhpDXzIAbG5KHW3mWOFIoxhynmU
# fln8jA/jb7UBJrZspe6HUSHkWGCbugwtK22ixH67xCUrRwIIfEmuE7bhfEJCKMYY
# Vs9BNLZmXbZ0e/VWMyIvIjayS6JKldj1po5SMIIFwDCCBKigAwIBAgIMA5bSC9ME
# D6d+m2gTMA0GCSqGSIb3DQEBCwUAMG4xCzAJBgNVBAYTAkJFMRkwFwYDVQQKExBH
# bG9iYWxTaWduIG52LXNhMUQwQgYDVQQDEztHbG9iYWxTaWduIEV4dGVuZGVkIFZh
# bGlkYXRpb24gQ29kZVNpZ25pbmcgQ0EgLSBTSEEyNTYgLSBHMzAeFw0xOTAyMjUw
# ODQzMTVaFw0yMjAzMjUwOTA3MjBaMIIBEzEdMBsGA1UEDwwUUHJpdmF0ZSBPcmdh
# bml6YXRpb24xEzARBgNVBAUTCkhSQiAxMDA1MjExEzARBgsrBgEEAYI3PAIBAxMC
# REUxHjAcBgsrBgEEAYI3PAIBAhMNTmllZGVyc2FjaHNlbjEdMBsGCysGAQQBgjc8
# AgEBEwxCcmF1bnNjaHdlaWcxCzAJBgNVBAYTAkRFMRYwFAYDVQQIEw1OaWVkZXJz
# YWNoc2VuMRIwEAYDVQQHEwlXb2xmc2J1cmcxHDAaBgNVBAkTE01heC12b24tTGF1
# ZS1XZWcgMTkxGDAWBgNVBAoTD1RFLVNZU1RFTVMgR21iSDEYMBYGA1UEAxMPVEUt
# U1lTVEVNUyBHbWJIMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA6WtX
# hHaDsjPr8MEBqii9Tw632GS3uxBgKzoFCZz706YzSFbEW8U7gQwLHttryRTo8zaG
# 4uRXAvItgFoFvxhASjehb8vfhFsWlAcik35MK2KkKAnLi30RouiheR0HTFtDwPx8
# iV76hmad9zR/589Cg0ETnNZb5atjhyC4O79d4i75PpLrDdX9itg2VWyNOLjFojP3
# 03neEl6SK52Wb2qVcfrkfWylUeStTOF/XOfeDZmfWCjOzzM2Iixf6hFRKfPmDPOu
# OziD1T/o31UPdHmrxTrgAaHCb383V5fhXHazjVCWrisZnloW7f4oA/gqdZPOKAn7
# AiSA4Ab+daBr500XgQIDAQABo4IBtTCCAbEwDgYDVR0PAQH/BAQDAgeAMIGgBggr
# BgEFBQcBAQSBkzCBkDBOBggrBgEFBQcwAoZCaHR0cDovL3NlY3VyZS5nbG9iYWxz
# aWduLmNvbS9jYWNlcnQvZ3NleHRlbmRjb2Rlc2lnbnNoYTJnM29jc3AuY3J0MD4G
# CCsGAQUFBzABhjJodHRwOi8vb2NzcDIuZ2xvYmFsc2lnbi5jb20vZ3NleHRlbmRj
# b2Rlc2lnbnNoYTJnMzBVBgNVHSAETjBMMEEGCSsGAQQBoDIBAjA0MDIGCCsGAQUF
# BwIBFiZodHRwczovL3d3dy5nbG9iYWxzaWduLmNvbS9yZXBvc2l0b3J5LzAHBgVn
# gQwBAzAJBgNVHRMEAjAAMEUGA1UdHwQ+MDwwOqA4oDaGNGh0dHA6Ly9jcmwuZ2xv
# YmFsc2lnbi5jb20vZ3NleHRlbmRjb2Rlc2lnbnNoYTJnMy5jcmwwEwYDVR0lBAww
# CgYIKwYBBQUHAwMwHQYDVR0OBBYEFJ10ZRvU0byT85jIVSrn1Gvv27MfMB8GA1Ud
# IwQYMBaAFNwsWCwqbzUtn3mVqEhdxG0+U7+5MA0GCSqGSIb3DQEBCwUAA4IBAQAJ
# Eno+CB8o4vvvK80fwWaTvb9wGxD2kdHW3rI64gsmicbnjFF7n9R9/XzdrowjXz0D
# 1UOxaf3JCPvqJNMOQvZJ1HHMebD1pcNouulUTkjsKczx4gG8GZjbIm3OwYWxVsWY
# tsDA3GdsP2ijKbEjuGpSL4LOIYY0S88yCpqyX+FpHPQRqXpkmWVlpHD8FCL/Ghar
# c5bNR8+HKDutCnEIH1dv3VGx6P8gerkZhMDGAaLW9QohP4/8l4kIXgMnJ8LfM3HO
# bMZv1Yc6aki3gB8AOdvvD8IMusn44nS5Rx4y56oZlVlN6z9CBVqiQVvCnd6/C9RC
# 5x6YhLvHjYejHrSts0G7MYIEUzCCBE8CAQEwfjBuMQswCQYDVQQGEwJCRTEZMBcG
# A1UEChMQR2xvYmFsU2lnbiBudi1zYTFEMEIGA1UEAxM7R2xvYmFsU2lnbiBFeHRl
# bmRlZCBWYWxpZGF0aW9uIENvZGVTaWduaW5nIENBIC0gU0hBMjU2IC0gRzMCDAOW
# 0gvTBA+nfptoEzAJBgUrDgMCGgUAoHgwGAYKKwYBBAGCNwIBDDEKMAigAoAAoQKA
# ADAZBgkqhkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgorBgEEAYI3AgELMQ4wDAYK
# KwYBBAGCNwIBFTAjBgkqhkiG9w0BCQQxFgQUsV9/ScVa8WGw/rdqXKALgNsujdkw
# DQYJKoZIhvcNAQEBBQAEggEAuoZmNEQibFW29NsgIFbk1NPks/HpW2iiRYmGQh69
# 4cW+s0ym05zNEQuPIO+4u4mrq89FEcFk9dNJbuifPJaO0cUsUAm9b/McGWX4FZJ+
# mZkzI1paejl8hySOdjs7OqsSCUJa5c+073yDaZ6hMecJooYWXcQpxtMu+6LFg19P
# 9rV9qFYotNkMmLurFMnLm8YM9By3WZMrf+CM2h5rgnk4/RkegNJS+xQRXnqOv4pQ
# dj3e9bkHJXpH/8j8t9fPv3HWcdUy0URE8ABjZ4I4T20CL9q6D9cP1t2oL7FhYhxA
# rRMPNbC7aq/1p8bk9P+nV8sITcTbIetQuSKE+/oS7qZABaGCAjAwggIsBgkqhkiG
# 9w0BCQYxggIdMIICGQIBATCBhjByMQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGln
# aUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNlcnQuY29tMTEwLwYDVQQDEyhE
# aWdpQ2VydCBTSEEyIEFzc3VyZWQgSUQgVGltZXN0YW1waW5nIENBAhANQkrgvjqI
# /2BAIc4UAPDdMA0GCWCGSAFlAwQCAQUAoGkwGAYJKoZIhvcNAQkDMQsGCSqGSIb3
# DQEHATAcBgkqhkiG9w0BCQUxDxcNMjEwNzA5MDgyMDU1WjAvBgkqhkiG9w0BCQQx
# IgQg9C6SSVZsMlmUsy6RQ2EmLgMKRvFrS8XFsJ14cYZD/1MwDQYJKoZIhvcNAQEB
# BQAEggEAj7yv0qCMnDf8G9/YeSP0MY3OP8cDrlyS9dmUEmaJA474G2+PkWYg32OP
# QJmBvowNwNiK2CV+0iXt+jvwuCgfZUuF3ZgioT4P7zMAVrG1KiTv4FsFrFZo1+hY
# fgKjD2OduPyE22CGw1E6H9Wz/vBX94Qan4dZNGJ2u48B3cz6jhKNOSoeBPH1lq9g
# 9m7oKdD+yAcGUfzZRri3nh45GyN+K6U5RA/IGE/DBjK1UNqVlXM6C6DHDdVRfEOo
# YBYkRTAOIWZBYlt1+vW0A3ux05//kc+nFqSrGmLpjizVyLkvfrQxsovXCI6u4vKc
# 9xkg9xUWTKimjzdZOluP3QUEkBo5JQ==
# SIG # End signature block
